use crate::numerics::float::Float;

// Fine structure constant.
pub const ALPHA: Float = 7.2973525693e-3;

// Barn unit, in cm^2.
pub const BARN: Float = 1e-24;

// Hartree energy unit, in MeV.
pub const HARTREE: Float = 27.211384523e-6;

// Avogadro number.
pub const AVOGADRO_NUMBER: Float = 6.02214085774e+23;

// Electron rest mass, in MeV.
pub const ELECTRON_MASS: Float = 0.5109989500015;

// Classical electron radius, in cm.
pub const ELECTRON_RADIUS: Float = 2.8179403262e-13;
